import unittest

from mock import call, patch

from gapipy.client import Client
from gapipy.request import APIRequestor

from .fixtures import FIRST_PAGE_LIST_DATA, SECOND_PAGE_LIST_DATA


@patch('gapipy.request.APIRequestor._request')
class APIRequestorTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()

    def test_get_resource_by_id(self, mock_request):
        requestor = APIRequestor(self.client, 'resources')
        requestor.get(1234)
        mock_request.assert_called_once_with('/resources/1234', 'GET')

    def test_list_resource(self, mock_request):
        requestor = APIRequestor(self.client, 'resources')
        requestor.list_raw()
        mock_request.assert_called_once_with('/resources', 'GET', options=None)

    def test_list_resource_with_parent(self, mock_request):
        parent = ('parent', '1234')
        requestor = APIRequestor(self.client, 'child', parent=parent)
        requestor.list_raw()
        mock_request.assert_called_once_with(
            '/parent/1234/child', 'GET', options=None)

    @patch('gapipy.request.APIRequestor.list_raw')
    def test_list_generator(self, mock_list, mock_request):
        mock_list.side_effect = [FIRST_PAGE_LIST_DATA, SECOND_PAGE_LIST_DATA]
        expected_calls = [
            call(None),
            call('http://localhost:5000/resources/?page=2'),
        ]

        requestor = APIRequestor(self.client, 'resource')
        resources = list(requestor.list())

        self.assertEqual(mock_list.mock_calls, expected_calls)
        self.assertEqual(len(resources), 6)



from importlib import import_module


def get_resource_class_from_class_name(name):
    resource_module = import_module('gapipy.resources')
    return getattr(resource_module, name)


def get_resource_class_from_resource_name(name):
    mapping = {resource._resource_name: resource
               for resource in get_available_resource_classes()}
    return mapping[name]


def get_available_resource_classes():
    from .resources import available_resources
    resource_module = import_module('gapipy.resources')
    return [getattr(resource_module, r) for r in available_resources]



# TODO: Replace fixtures by factories (or anything more manageable)
# TODO: Replace hardcoded urls with config.API_ROOT?

PPP_TOUR_DATA = {
    'departures': {'href': 'http://localhost:5000/tours/21346/departures'},
    'departures_end_date': '2013-12-31',
    'departures_start_date': '2013-01-01',
    'tour_dossier': {
        'href': 'http://localhost:5000/tour_dossiers/21346',
        'id': '21346'
    },
    'product_line': 'PPP',
    'href': 'http://localhost:5000/tours/21346',
    'id': '21346'
}


TOUR_DOSSIER_LIST_DATA = {
    'count': 3,
    'max_per_page': 3,
    'current_page': 1,
    'results': [
        {
            'href': 'http://localhost:5000/tour_dossiers/1234',
            'id': '1234',
            'tour_dossier_code': 'ABCD'
        },
        {
            'href': 'http://localhost:5000/tour_dossiers/5678',
            'id': '5678',
            'tour_dossier_code': 'EFGH'
        },
        {
            'href': 'http://localhost:5000/tour_dossiers/9012',
            'id': '9012',
            'tour_dossier_code': 'IJKL'
        }
    ],
    'links': []
}

FIRST_PAGE_LIST_DATA = {
    'count': 6,
    'max_per_page': 3,
    'current_page': 1,
    'results': [
        {
            'href': 'http://localhost:5000/resources/1234',
            'id': '1234',
        },
        {
            'href': 'http://localhost:5000/resources/5678',
            'id': '5678',
        },
        {
            'href': 'http://localhost:5000/resources/9012',
            'id': '9012',
        }
    ],
    'links': [
        {
            'href': 'http://localhost:5000/resources/?page=2',
            'rel': 'next',
        }
    ]
}

SECOND_PAGE_LIST_DATA = {
    'count': 6,
    'max_per_page': 3,
    'current_page': 2,
    'results': [
        {
            'href': 'http://localhost:5000/resources/3456',
            'id': '3456',
        },
        {
            'href': 'http://localhost:5000/resources/7890',
            'id': '7890',
        },
        {
            'href': 'http://localhost:5000/resources/1111',
            'id': '1111',
        }
    ],
    'links': [
        {
            'href': 'http://localhost:5000/resources/',
            'rel': 'prev',
        }
    ]
}
PPP_DOSSIER_DATA = {
    "departures_start_date": "2013-01-01",
    "departures_end_date": "2013-12-31",
    "itineraries": [
        {
            "duration": 15,
            "type": "SUMMARY",
            "days": [
                {
                    "body": "Arrive at any time.",
                    "id": "125737468",
                    "label": "Day 1 Lima"
                },
                {
                    "body": "Fly to Juliaca and transfer to Puno.  Visit the floating Islands of Uros and take a guided tour of Lake Titicaca with a homestay in a small village.   Optional visit to Sillustani burial site.",
                    "id": "125737469",
                    "label": "Days 2-4 Puno/Lake Titicaca (1B,1L,1D)"
                },
                {
                    "body": "Travel day by bus from Puno to Cuzco.",
                    "id": "125737470",
                    "label": "Day 5 Cuzco"
                },
                {
                    "body": "Full day guided tour of the Sacred Valley and Ollantaytambo ruins along with a visit to a Planeterra weaving project in a local community. Shopping opportunities at a local market.",
                    "id": "125737471",
                    "label": "Day 6 Sacred Valley/Ollantaytambo"
                },
                {
                    "body": "4-day guided Inca Trail hike with visit to Machu Picchu. Optional visit to the Inca Bridge before returning to Cuzco.",
                    "id": "125737472",
                    "label": "Days 7-10 Inca Trail/Machu Picchu (3B,3L,3D)"
                },
                {
                    "body": "Free day to explore Cuzco or relax. Active options include whitewater rafting, horseback riding and mountain biking.",
                    "id": "125737473",
                    "label": "Day 11 Cuzco"
                },
                {
                    "body": "Fly to Puerto Maldonado and continue by motorized canoe to our comfortable, intimate and exclusive G Lodge Amazon. Enjoy guided excursions by expert naturalists to spot wildlife at nearby oxbow lakes, clay licks and treetop towers. Included rubber boots while at the lodge. Fly back to Lima on Day 14.",
                    "id": "125737474",
                    "label": "Days 12-14 Amazon Jungle (2B,2L,2D)"
                },
                {
                    "body": "Depart at any time.",
                    "id": "125737475",
                    "label": "Day 15 Lima"
                }
            ]
        }
    ],
    "description": "Spot macaws in the jungle and caimans on the riverbanks, sail the waters of Lake Titicaca, delight in the smells of markets and explore ancient ruins, including a trek along the Inca Trail. As we operate our own treks, our quality equipment and the expertise of our porters and CEOs will ensure that your first glimpse of Machu Picchu will leave you in awe. Whether you're scanning the canopy for wildlife from the comfort of our intimate and exclusive G Lodge Amazon or following in the footsteps of the Incas, this fast-paced adventure provides a sweeping view of this diverse country.",
    "tour": {
        "href": "http://localhost:5000/tours/21346",
        "id": "21346"
    },
    "href": "http://localhost:5000/tour_dossiers/21346",
    "product_line": "PPP",
    "images": [
        {
            "image_href": "https://media.gadventures.com/media-server/dynamic/admin/maps/2013/PPP.jpg",
            "type": "MAP"
        },
        {
            "image_href": "https://media.gadventures.com/media-server/dynamic/admin/trip-banners/south-america/peru/PPP/PPP_banner_peru_landscape_ruins_machu_picchu_clouds.jpg",
            "type": "BANNER"
        }
    ],
    "details": [
        {
            "body": "Meeting the locals at a Lake Titicaca homestay, climbing the Inca Trail to Machu Picchu, spotting wildlife at our exclusive G Lodge Amazon, devouring seafood in Lima.",
            "detail_type": {
                "id": "2",
                "label": "Highlights"
            }
        },
        {
            "body": "16",
            "detail_type": {
                "id": "113",
                "label": "Max Pax"
            }
        },
        {
            "body": "Chief Experience Officer (CEO) throughout, specialist Inca Trail CEO on hike.",
            "detail_type": {
                "id": "18",
                "label": "Group Leader"
            }
        }
    ],
    "geography": {
        "start_city": {
            "name": "Lima"
        },
        "start_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "region": {
            "name": "South America"
        },
        "visited_countries": [
            {
                "href": "http://localhost:5000/countries/PE",
                "id": "PE",
                "name": "Peru"
            }
        ],
        "finish_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "primary_country": {
            "href": "http://localhost:5000/countries/PE",
            "id": "PE",
            "name": "Peru"
        },
        "finish_city": {
            "name": "Lima"
        }
    },
    "id": "21346",
    "categories": [
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Wildlife/Nature"
        },
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Trekking/Hiking"
        },
        {
            "category_type": {
                "id": "1",
                "label": "Activity"
            },
            "name": "Culture/History"
        },
        {
            "category_type": {
                "id": "16",
                "label": "Trip Style"
            },
            "name": "Classic"
        },
        {
            "category_type": {
                "id": "28",
                "label": "Service Level"
            },
            "name": "Standard"
        },
        {
            "category_type": {
                "id": "33",
                "label": "Physical Grading"
            },
            "name": "4"
        },
        {
            "category_type": {
                "id": "39",
                "label": "Merchandising"
            },
            "name": "Top Seller"
        },
        {
            "category_type": {
                "id": "58",
                "label": "Trip Type"
            },
            "name": "Small Group"
        }
    ],
    "name": "Peru Panorama"
}


PRODUCTS_DATA = [
    {
        "href": "https://localhost:5000/departures/1234",
        "id": "1234",
        "type": "departures",
        "sub_type": "Tour",
    },
    {
        "href": "https://localhost:5000/departures/5678",
        "id": "5678",
        "type": "departures",
        "sub_type": "Tour",
    },
    {
        "href": "https://localhost:5000/departures/9012",
        "id": "9012",
        "type": "departures",
        "sub_type": "Tour",
    }
]

DUMMY_DEPARTURE = {
    "sku": "GAPPPP130406-O1",
    "product_line": "PPP",
    "tour_dossier": {
        "href": "https://rest.gadventures.com/tour_dossiers/9882",
        "id": "9882",
        "name": "Peru Panorama"
    },
    "date_last_modified": "2013-06-10T19:12:04Z",
    "finish_date": "2013-04-20",
    "lowest_pp2a_prices": [
        {
            "currency": "USD",
            "amount": "3099.00"
        },
        {
            "currency": "AUD",
            "amount": "2999.00"
        },
        {
            "currency": "CHF",
            "amount": "2949.00"
        },
        {
            "currency": "GBP",
            "amount": "1749.00"
        },
        {
            "currency": "NZD",
            "amount": "3999.00"
        },
        {
            "currency": "CAD",
            "amount": "3299.00"
        },
        {
            "currency": "EUR",
            "amount": "2149.00"
        }
    ],
    "availability": {
        "status": "NOT_BOOKABLE",
        "total": 8
    },
    "tour": {
        "href": "https://rest.gadventures.com/tours/9882",
        "id": "9882"
    },
    "nearest_finish_airport": {
        "code": "LIM"
    },
    "id": "419504",
    "latest_arrival_time": "2013-04-06T23:59:59",
    "href": "https://rest.gadventures.com/departures/419504",
    "flags": [],
    "earliest_departure_time": "2013-04-20T00:00:00",
    "add_ons": [
        {
            "name": "San Agustin Colonial",
            "finish_date": "2013-04-06",
            "min_days": 1,
            "id": "54",
            "href": "https://rest.gadventures.com/accommodations/54",
            "max_days": 5,
            "start_date": "2013-04-02",
            "type": "accommodations",
            "sub_type": "Hotel"
        },
        {
            "name": "San Agustin Colonial",
            "finish_date": "2013-04-24",
            "min_days": 1,
            "id": "54",
            "href": "https://rest.gadventures.com/accommodations/54",
            "max_days": 5,
            "start_date": "2013-04-20",
            "type": "accommodations",
            "sub_type": "Hotel"
        },
        {
            "name": "Airport to Hotel Transfer",
            "finish_date": "2013-04-06",
            "min_days": 1,
            "id": "298",
            "href": "https://rest.gadventures.com/transports/298",
            "max_days": 1,
            "start_date": "2013-04-02",
            "type": "transports",
            "sub_type": "Transfer"
        },
        {
            "name": "Peru Culinary Theme Pack",
            "finish_date": "2013-04-17",
            "min_days": 1,
            "id": "4486",
            "href": "https://rest.gadventures.com/activities/4486",
            "max_days": 1,
            "start_date": "2013-04-16",
            "type": "activities",
            "sub_type": "Product"
        }
    ],
    "date_created": "2012-01-26T20:29:07Z",
    "nearest_start_airport": {
        "code": "LIM"
    },
    "start_date": "2013-04-06",
    "rooms": [
        {
            "code": "STANDARD",
            "name": "Standard",
            "flags": [],
            "price_bands": [
                {
                    "max_age": 100,
                    "code": "ADULT",
                    "name": "Adult",
                    "max_travellers": 16,
                    "min_age": 12,
                    "prices": [
                        {
                            "promotions": [],
                            "currency": "AUD",
                            "amount": "2999.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [],
                            "currency": "CAD",
                            "amount": "3299.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [],
                            "currency": "CHF",
                            "amount": "2949.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [],
                            "currency": "EUR",
                            "amount": "2149.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [],
                            "currency": "GBP",
                            "amount": "1749.00",
                            "deposit": "100.00"
                        },
                        {
                            "promotions": [],
                            "currency": "NZD",
                            "amount": "3999.00",
                            "deposit": "250.00"
                        },
                        {
                            "promotions": [],
                            "currency": "USD",
                            "amount": "3099.00",
                            "deposit": "250.00"
                        }
                    ],
                    "min_travellers": 1
                }
            ],
            "add_ons": [
                {
                    "name": "My Own Room",
                    "finish_date": "2013-04-20",
                    "min_days": 14,
                    "id": "T419504",
                    "href": "https://rest.gadventures.com/single_supplements/T419504",
                    "max_days": 14,
                    "start_date": "2013-04-06",
                    "type": "single_supplements",
                    "sub_type": "Single Supplement"
                }
            ],
            "availability": {
                "status": "NOT_BOOKABLE",
                "total": 8,
                "male": None,
                "female": None
            }
        }
    ]
}


DUMMY_PROMOTION = {
    "currencies": ["AUD", "CAD", "CHF", "EUR", "GBP", "NZD", "USD"],
    "name": "10% off North America",
    "commission_rate": 0,
    "sale_finish_date": "2012-08-27",
    "discount_percent": "10.00",
    "id": "18336",
    "terms_and_conditions": "From big cities to untamed wildlife and scenic landscapes, there's a North American adventure for everyone.\r\n<ul><li>Travellers receive 10% off all North America tours (valid for all North America trip codes).</li>\r\n<li>Book by August 27, 2012 for departures through December 16, 2012.</li></ul>\r\nFor general Terms and Conditions please visit our <a href=http://www.gadventures.com/terms-and-conditions/>Terms and Conditions page.</a>",
    "product_finish_date": None,
    "discount_amount": None,
    "href": "https://localhost:5000/promotions/18336",
    "sale_start_date": "2012-08-14",
    "product_start_date": None,
    "room_codes": None,
    "flags": None,
    "products": PRODUCTS_DATA,
    "promotion_code": "BZ-NOAM10",
    "min_travellers": 0
}



from __future__ import unicode_literals

from ..base import Resource


class Nationality(Resource):

    _resource_name = 'nationalities'

    _as_is_fields = ['id', 'href', 'name']

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)



import requests

from . import __title__, __version__


ACCEPTABLE_RESPONSE_STATUS_CODES = (
    requests.codes.ok, requests.codes.created, requests.codes.accepted,
)


class APIRequestor(object):

    def __init__(self, client, resource, options=None, parent=None):
        self.client = client
        self.resource = resource
        self.options = options
        self.parent = parent

    def _request(self, uri, method, data=None, options=None, additional_headers=None):
        """Make an HTTP request to a target API method with proper headers."""

        assert method in ['GET', 'POST', 'PUT', 'PATCH'], "Only 'GET', 'POST', 'PUT', and 'PATCH' are allowed."

        # Support supplying a full url
        if '://' in uri:
            url = uri
        else:
            url = self.client.api_root + uri

        # Strip out the proxy from the url. The client only wants to return urls
        # with the API_PROXY, but not actually query on them.
        api_proxy = self.client.api_proxy
        if api_proxy:
            url = url.replace(api_proxy, '')

        headers = {
            'Content-Type': 'application/json',
            'User-Agent': '{0}/{1}'.format(__title__, __version__),
            'X-Application-Key': self.client.application_key,
            'Accept-Language': self.client.api_language,
        }
        if additional_headers:
            headers.update(additional_headers)

        if api_proxy:
            headers.update({'X-Api-Proxy': api_proxy})

        requests_call = getattr(requests, method.lower())

        self.client.logger.debug('Making a {0} request to {1}'.format(method, url))
        response = requests_call(url, headers=headers, data=data, params=options)

        if response.status_code in ACCEPTABLE_RESPONSE_STATUS_CODES:
            return response.json()
        else:
            response.reason = response.text
            return response.raise_for_status()

    def get(self, resource_id=None, uri=None):
        """Get a single resource with the given resource_id or uri"""

        if not (resource_id or uri):
            raise ValueError(
                'Need to provide at least one of `resource_id` or `uri` as argument')
        if not uri:
            uri = '/{0}/{1}'.format(self.resource, resource_id)
        return self._request(uri, 'GET')

    def update(self, resource_id, data, partial=True, uri=None):
        """
        Update a single resource with the given data.

        For GAPI, all updates are done using the `PATCH` method, regardless of the `partial` argument.
        Well, that's the plan, once the charming load balancer is updated.
        """
        method = 'PUT'
        if not uri:
            uri = '/{0}/{1}'.format(self.resource, resource_id)
        return self._request(uri, method, data=data, additional_headers={'X-HTTP-Method-Override': 'PATCH'})

    def create(self, data, uri=None):
        """
        Create a single new resource with the given data.
        """
        if not uri:
            uri = '/{0}'.format(self.resource)
        return self._request(uri, 'POST', data=data)

    def list_raw(self, uri=None):
        """Return the raw response for listing resources.

        If `uri` is given, make a GET request to it, otherwise build the uri
        from `self.resource` and `self.parent`. Note that only the first page
        of results will be returned. To get an iterator over all resources,
        use `list`.
        """
        if not uri:
            if self.parent:
                parent_name, parent_id = self.parent
                uri = '/{0}/{1}/{2}'.format(parent_name, parent_id, self.resource)
            else:
                uri = '/{0}'.format(self.resource)

        return self._request(uri, 'GET', options=self.options)

    def list(self, uri=None):
        """Generator for listing resources"""

        response = self.list_raw(uri)
        for result in response['results']:
            yield result

        for link in response.get('links', []):
            if link['rel'] != 'next':
                continue
            for result in self.list(link['href']):
                yield result



from __future__ import unicode_literals

from ..base import Resource


class Document(Resource):
    _resource_name = 'documents'
    _is_listable = False

    _as_is_fields = ['id', 'href', 'mime_type', 'content', 'type']
    _date_time_fields_utc = ['date_created']
    _resource_fields = [
        ('booking', 'Booking'),
    ]


class Invoice(Resource):
    _resource_name = 'invoices'
    _is_listable = False

    _as_is_fields = ['id', 'href']
    _date_time_fields_utc = ['date_created']
    _resource_fields = [
        ('document', Document),
        ('booking', 'Booking'),
    ]



from .base import BaseModel


class DepartureFlightDetail(BaseModel):
    _as_is_fields = ['departure_flight_number']
    _date_time_fields_local = ['departure_flight_date_time']

    @property
    def _resource_fields(self):
        from gapipy.resources import Customer
        return [
            ('customer', Customer),
        ]



from .base import BaseModel, RelatedResourceMixin


class AddOn(BaseModel, RelatedResourceMixin):
    _as_is_fields = [
        'id', 'href', 'max_days', 'min_days', 'name', 'type', 'sub_type'
    ]
    _date_fields = ['start_date', 'finish_date']
    _related_resource_lookup = 'type'

    def __repr__(self):
        return '<{0} ({1})>'.format(self.__class__.__name__, self.name)



# flake8: NOQA
from .accommodation import Accommodation
from .activity import Activity
from .departure import Departure
from .promotion import Promotion
from .single_supplement import SingleSupplement
from .tour import Tour
from .tour_category import TourCategory
from .tour_dossier import TourDossier
from .transport import Transport



from .base import BaseModel


class DocumentInfo(BaseModel):
    """Represents a document, without its content"""
    _as_is_fields = ['id', 'href', 'mime_type', 'type']
    _date_time_fields_utc = ['date_created']
    _resource_fields = [
        ('booking', 'Booking'),
    ]



# flake8: NOQA
from .add_on import AddOn
from .address import Address
from .price_band import PriceBand, SeasonalPriceBand, PP2aPrice
from .room import AccommodationRoom, DepartureRoom
from .incomplete_requirement import IncompleteRequirement
from .arrival_flight_detail import ArrivalFlightDetail
from .departure_flight_detail import DepartureFlightDetail
from .international_ticket_number import InternationalTicketNumber
from .traveller_height import TravellerHeight
from .departure_service_room import DepartureServiceRoom
from .document_info import DocumentInfo

from .base import DATE_FORMAT, DATE_TIME_UTC_FORMAT, DATE_TIME_LOCAL_FORMAT



from .tour import (  # NOQA
    Accommodation,
    Activity,
    Departure,
    Promotion,
    SingleSupplement,
    Tour,
    TourCategory,
    TourDossier,
    Transport,
)

from .booking import (  # NOQA
    Agency,
    Agent,
    Booking,
    Customer,
    Invoice,
    Document,
    Nationality,
    AccommodationService,
    ActivityService,
    DepartureService,
    FeeService,
    FlightService,
    InsuranceService,
    SingleSupplementService,
    TransportService,
    Payment,
    Refund,
)

from .geo import (  # NOQA
    Airport,
    Continent,
    Country,
    Feature,
    FeatureCategory,
    Place,
    State,
    Timezone,
)

from .language import Language  # NOQA


available_public_resources = [
    # Tour
    'Accommodation', 'Activity', 'Departure', 'Promotion', 'SingleSupplement',
    'Tour', 'TourCategory', 'TourDossier', 'Transport',

    # Geographical
    'Airport', 'Continent', 'Country', 'Feature', 'FeatureCategory', 'Place',
    'State', 'Timezone',

    'Language'
]

available_private_resources = [
    'Agency', 'Agent', 'Booking', 'Customer', 'Invoice', 'Document',
    'Nationality', 'AccommodationService', 'ActivityService',
    'DepartureService', 'FeeService', 'FlightService', 'InsuranceService',
    'SingleSupplementService', 'TransportService', 'Payment', 'Refund',
]

available_resources = available_public_resources + available_private_resources



from __future__ import unicode_literals

from ..base import Resource


class State(Resource):

    _resource_name = 'states'

    _as_is_fields = ['id', 'href', 'name']
    _resource_fields = [('country', 'Country')]

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)



from functools import wraps
from itertools import islice

from .request import APIRequestor


def _check_listable(func):

    @wraps(func)
    def inner(query, *args, **kwargs):
        if not (query.resource._is_listable or query.parent):
            raise ValueError(
                'The {0} resource is not listable and/or is only available as a subresource'.format(
                    query.resource.__name__))

        return func(query, *args, **kwargs)

    return inner


class Query(object):

    def __init__(self, client, resource, filters=None, parent=None, raw_data=None):
        self._client = client
        self._raw_data = raw_data or {}
        self.resource = resource
        self.filters = filters or {}
        self.parent = parent

    def _to_dict(self):
        # Used by Resource when converting nested Query objects into
        # serializable types.
        return self._raw_data

    def get(self, resource_id):
        """Returns an instance of the query resource with the given resource_id."""

        resource_data = self._client._cache.get(self.resource._resource_name, resource_id)
        if resource_data is not None:
            return self.resource(resource_data)

        requestor = APIRequestor(self._client, self.resource._resource_name)
        resource_object = self.resource(requestor.get(resource_id))
        self._client._cache.set(self.resource._resource_name, resource_object.to_dict())
        return resource_object

    @_check_listable
    def all(self, limit=None):
        """Generator of instances of the query resource. If limit is set to a
        positive integer `n`, only return the first `n` results.
        """

        requestor = APIRequestor(
            self._client,
            self.resource._resource_name,
            options=self.filters,
            parent=self.parent
        )
        generator = requestor.list()

        if limit:
            if isinstance(limit, int) and limit > 0:
                generator = islice(generator, limit)
            else:
                raise ValueError('`limit` must be a positive integer')

        for result in generator:
            yield self.resource(result, stub=True)

    def filter(self, **kwargs):
        """Add filter arguments to the query.

        For example, if `query` is a Query for the TourDossier ressource, then
        `query.filter(name='Amazing Adventure')` will return a query containing
        only dossiers whose names contain 'Amazing Adventure'.
        """

        self.filters.update(kwargs)
        return self

    @_check_listable
    def count(self):
        """Returns the number of element in the query."""

        requestor = APIRequestor(
            self._client,
            self.resource._resource_name,
            options=self.filters,
            parent=self.parent
        )
        response = requestor.list_raw()
        return response.get('count')

    def __iter__(self):
        """Provided as a convenience so that Query objects can be iterated
        without calling `all`.

            i.e.  `for dossier in dossiers.filter(name='Peru')`
                instead of `for dossier in dossiers.filter(name='Peru').all()`
        """
        return self.all()



from __future__ import unicode_literals

from ...utils import get_resource_class_from_resource_name
from ...models import (
    IncompleteRequirement, ArrivalFlightDetail, DepartureFlightDetail,
    InternationalTicketNumber, TravellerHeight, DepartureServiceRoom,
    DocumentInfo,
)

from ..base import Resource

from ..tour import (
    Departure,
    Accommodation,
    Activity,
    Transport,
    SingleSupplement,
)
from .customer import Customer


class TypeBasedServiceMeta(type):
    def __call__(cls, *args, **kwargs):
        # Attempt to extract the type from the data. If it's not found for
        # whatever reason, we can't assume what the user was attempting, and
        # we'll simply ignore any metaclass magic.
        resource_name = args[0].get('type', None) if args else None

        if resource_name is None:
            return type.__call__(cls, *args, **kwargs)

        new_class = get_resource_class_from_resource_name(resource_name)
        return type.__call__(new_class, *args, **kwargs)


class Service(Resource):
    _resource_name = 'services'
    _is_listable = False
    _is_parent_resource = True

    __metaclass__ = TypeBasedServiceMeta

    @property
    def _as_is_fields(self):
        return [
            'id', 'href', 'name', 'status', 'type', 'sub_type', 'flags',
            'applied_promotion',
            'status_transitions',
        ]

    @property
    def _date_fields(self):
        return ['start_date', 'finish_date']

    @property
    def _date_time_fields_utc(self):
        return [
            'date_created', 'date_confirmed',
            'date_cancelled', 'option_expiry_date',
        ]

    @property
    def _date_time_fields_local(self):
        return []

    @property
    def _price_fields(self):
        return [
            'purchase_price', 'commission',
        ]

    @property
    def _resource_fields(self):
        return [
            ('booking', 'Booking'),
            ('customers', Customer),
        ]

    @property
    def _model_collection_fields(self):
        return [
            ('documents', DocumentInfo),
        ]


class ServiceProduct(Service):
    def _get_product(self):
        # Return the appropriate product based on the parent service.
        mapping = {
            'departure_services': Departure,
            'accommodation_services': Accommodation,
            'transport_services': Transport,
            'activity_services': Activity,
            'single_supplement_services': SingleSupplement,
        }
        return mapping[self._resource_name]

    @property
    def _resource_fields(self):
        return super(ServiceProduct, self)._resource_fields + [
            ('product', self._get_product()),
        ]


class DepartureService(ServiceProduct):
    _resource_name = 'departure_services'

    @property
    def _model_collection_fields(self):
        return (super(DepartureService, self)._model_collection_fields + [
            ('rooms', DepartureServiceRoom),
            ('incomplete_requirements', IncompleteRequirement),
            ('arrival_flight_details', ArrivalFlightDetail),
            ('departure_flight_details', DepartureFlightDetail),
            ('international_ticket_numbers', InternationalTicketNumber),
            ('traveller_heights', TravellerHeight),
        ])


class AccommodationService(ServiceProduct):
    _resource_name = 'accommodation_services'

    @property
    def _as_is_fields(self):
        return super(AccommodationService, self)._as_is_fields + [
            'room',
        ]


class TransportService(ServiceProduct):
    _resource_name = 'transport_services'

    @property
    def _as_is_fields(self):
        return super(TransportService, self)._as_is_fields + [
            'flight_number', 'pickup_time',
        ]


class ActivityService(ServiceProduct):
    _resource_name = 'activity_services'

    @property
    def _model_collection_fields(self):
        return (super(ActivityService, self)._model_collection_fields + [
            ('incomplete_requirements', IncompleteRequirement),
            ('arrival_flight_details', ArrivalFlightDetail),
        ])


class SingleSupplementService(ServiceProduct):
    _resource_name = 'single_supplement_services'


class InsuranceService(Service):
    _resource_name = 'insurance_services'

    @property
    def _as_is_fields(self):
        return super(InsuranceService, self)._as_is_fields + [
            'policy_type',
            'policy_number',
            'policy_details_url',
        ]

    @property
    def _price_fields(self):
        return super(InsuranceService, self)._price_fields + [
            'policy_amount',
            'insurable_amount',
        ]


class FeeService(Service):
    _resource_name = 'fee_services'

    @property
    def _as_is_fields(self):
        return super(FeeService, self)._as_is_fields + [
            'description',
            'related_service',
        ]


class FlightService(Service):
    _resource_name = 'flight_services'

    @property
    def _as_is_fields(self):
        return super(FlightService, self)._as_is_fields + [
            'record_locator',
            'itinerary_url',
            'segments',
        ]



from __future__ import unicode_literals

from ...utils import (
    get_resource_class_from_class_name, get_resource_class_from_resource_name
)
from ..base import Resource, Product


class Promotion(Resource):

    _resource_name = 'promotions'

    _as_is_fields = [
        'id', 'href', 'commission_rate', 'currencies', 'flags',
        'min_travellers', 'name', 'promotion_code', 'room_codes',
        'terms_and_conditions',
    ]

    _date_fields = [
        'product_start_date', 'product_finish_date',
        'sale_start_date', 'sale_finish_date'
    ]
    _price_fields = ['discount_amount', 'discount_percent']
    _resource_collection_fields = [('products', Product)]

    def _fill_resource_collection_fields(self, data):
        for field, resource_cls in self._resource_collection_fields:
            if isinstance(resource_cls, basestring):
                resource_cls = get_resource_class_from_class_name(resource_cls)

            products = data[field]
            if products:
                product_type = products[0]['type']
                resource_cls = get_resource_class_from_resource_name(product_type)
                product_stubs = [resource_cls(p, stub=True) for p in products]
                setattr(self, field, product_stubs)
            else:
                setattr(self, field, [])



import json
import unittest

from mock import patch

from gapipy.client import Client
from gapipy.query import Query
from gapipy.resources import Departure, Tour, TourDossier
from gapipy.resources.base import Resource

from .fixtures import (
    PPP_TOUR_DATA, TOUR_DOSSIER_LIST_DATA, DUMMY_DEPARTURE, DUMMY_PROMOTION,
)


class QueryTestCase(unittest.TestCase):

    def setUp(self):
        self.client = Client()
        self.cache = self.client._cache
        self.cache.clear()

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_get_instance_by_id(self, mock_request):
        query = Query(self.client, Tour)
        t = query.get(1234)
        self.assertIsInstance(t, Tour)

    @patch('gapipy.request.APIRequestor._request', return_value=PPP_TOUR_DATA)
    def test_resources_are_cached(self, mock_request):
        query = Query(self.client, Tour)
        self.assertEqual(self.cache.count(), 0)

        query.get(21346)
        self.assertEqual(self.cache.count(), 1)
        self.assertEqual(len(mock_request.mock_calls), 1)

        query.get(21346)
        self.assertEqual(len(mock_request.mock_calls), 1)

    @patch('gapipy.request.APIRequestor._request')
    def test_filtered_query(self, mock_request):
        query = Query(self.client, Tour).filter(tour_dossier_code='PPP')
        list(query)  # force query evaluation
        mock_request.assert_called_once_with(
            '/tours', 'GET', options={'tour_dossier_code': 'PPP'})

        # Check that filters can be chained
        list(query.filter(departures_start_date='2014-01-01'))
        mock_request.assert_called_with(
            '/tours', 'GET', options={
                'tour_dossier_code': 'PPP',
                'departures_start_date': '2014-01-01'
            })

    def test_listing_non_listable_resource_fails(self):
        message = 'The Departure resource is not listable and/or is only available as a subresource'
        with self.assertRaisesRegexp(ValueError, message):
            Query(self.client, Departure).all()
        with self.assertRaisesRegexp(ValueError, message):
            Query(self.client, Departure).count()

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_PROMOTION)
    def test_can_retrieve_single_non_listable_resource(self, mock_request):
        Query(self.client, Departure).get(1234)
        mock_request.assert_called_once_with(
            '/departures/1234', 'GET')

    @patch('gapipy.request.APIRequestor._request', return_value=DUMMY_DEPARTURE)
    def test_can_retrieve_single_subresource_without_parent(self, mock_request):
        Query(self.client, Departure).get(1234)
        mock_request.assert_called_once_with(
            '/departures/1234', 'GET')

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_count(self, mock_request):
        query = Query(self.client, TourDossier)
        count = query.count()
        self.assertIsInstance(count, int)
        self.assertEqual(count, 3)

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all(self, mock_request):

        query = Query(self.client, TourDossier).all()
        self.assertFalse(mock_request.called)

        dossiers = list(query)
        self.assertTrue(mock_request.called)

        self.assertEqual(len(dossiers), 3)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', options={})

    @patch('gapipy.request.APIRequestor._request', return_value=TOUR_DOSSIER_LIST_DATA)
    def test_fetch_all_with_limit(self, mock_request):

        query = Query(self.client, TourDossier).all(limit=2)
        dossiers = list(query)

        self.assertEqual(len(dossiers), 2)
        for dossier in dossiers:
            self.assertIsInstance(dossier, TourDossier)

        mock_request.assert_called_once_with(
            '/tour_dossiers', 'GET', options={})

    def test_fetch_all_with_wrong_argument_for_limit(self):
        message = '`limit` must be a positive integer'
        with self.assertRaisesRegexp(ValueError, message):
            query = Query(self.client, Tour).all(limit=-1)
            list(query)  # force the query to evaluate


class MockResource(Resource):
    _as_is_fields = ['id', 'first_name', 'last_name']
    _resource_name = 'mocks'


@patch('gapipy.request.APIRequestor._request')
class UpdateCreateResourceTestCase(unittest.TestCase):
    def test_object_accessor(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data)
        self.assertEquals(r.first_name, data['first_name'])

    def test_object_attr_modify(self, mock_request):
        data = {'first_name': 'Jon', 'last_name': 'Ive', 'id': None}
        r = MockResource(data)
        r.first_name = 'Jonathan'
        self.assertEquals(r.first_name, 'Jonathan')

    def test_create_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': None,
        }
        r = MockResource(data)
        r.save()
        mock_request.assert_called_once_with(
            '/mocks', 'POST', data=json.dumps(data))

    def test_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        r = MockResource(data)

        r.first_name = 'Jonathan'
        r.save()
        data['first_name'] = 'Jonathan'

        mock_request.assert_called_once_with(
            '/mocks/1', 'PUT', data=json.dumps(data), additional_headers={'X-HTTP-Method-Override': 'PATCH'})

    def test_partial_update_object(self, mock_request):
        data = {
            'first_name': 'Jon',
            'last_name': 'Ive',
            'id': 1,
        }
        mock_request.return_value = data
        r = MockResource(data)

        r.first_name = 'Jonathan'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ive',
            'id': 1
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'first_name': 'Jonathan'}
        mock_request.assert_called_once_with(
            '/mocks/1', 'PUT', data=json.dumps(changed), additional_headers={'X-HTTP-Method-Override': 'PATCH'})
        self.assertEquals(r.to_dict(), r_data)

        r.last_name = 'Ivey'
        r_data = {
            'first_name': 'Jonathan',
            'last_name': 'Ivey',
            'id': 1,
        }
        mock_request.return_value = r_data
        r.save(partial=True)
        changed = {'last_name': 'Ivey'}
        mock_request.assert_called_with(
            '/mocks/1', 'PUT', data=json.dumps(changed), additional_headers={'X-HTTP-Method-Override': 'PATCH'})
        self.assertEquals(r.to_dict(), r_data)



import logging
import os
from importlib import import_module

from .utils import get_available_resource_classes


current_client = None

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())

default_config = {
    'application_key': os.environ.get('GAPI_APPLICATION_KEY'),
    'api_root': os.environ.get('GAPI_API_ROOT', 'https://rest.gadventures.com'),
    'api_proxy': os.environ.get('GAPI_API_PROXY', ''),
    'api_language': os.environ.get('GAPI_LANGUAGE', 'en'),
    'cache_backend': os.environ.get('GAPI_CACHE_BACKEND', 'gapipy.cache.SimpleCache'),
    'debug': os.environ.get('GAPI_CLIENT_DEBUG', False),
}


def get_config(config, name):
    return config.get(name, default_config[name])


class Client(object):

    def __init__(self, **config):

        global current_client
        current_client = self

        self.application_key = get_config(config, 'application_key')
        self.api_root = get_config(config, 'api_root')
        self.api_proxy = get_config(config, 'api_proxy')
        self.api_language = get_config(config, 'api_language')
        self.cache_backend = get_config(config, 'cache_backend')

        log_level = 'DEBUG' if get_config(config, 'debug') else 'ERROR'
        self.logger = logger
        self.logger.setLevel(log_level)

        self._set_cache_instance()

        # Prevent install issues where setup.py digs down the path and
        # eventually fails on a missing requests requirement by importing Query
        # only where it's needed.
        from .query import Query
        for resource in get_available_resource_classes():
            setattr(self, resource._resource_name, Query(self, resource))

    def _set_cache_instance(self):
        cache_backend = self.cache_backend
        module_name, class_name = cache_backend.rsplit('.', 1)
        module = import_module(module_name)
        cache = getattr(module, class_name)()
        self._cache = cache

    def query(self, resource_name):
        try:
            return getattr(self, resource_name)
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

    def build(self, resource_name, data_dict, **kwargs):
        try:
            resource_cls = getattr(self, resource_name).resource
            return resource_cls(data_dict, **kwargs)
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

    def create(self, resource_name, data_dict):
        """
        Create an instance of the specified resource with `data_dict`
        """
        try:
            resource_cls = getattr(self, resource_name).resource
        except AttributeError:
            raise AttributeError("No resource named %s is defined." % resource_name)

        return resource_cls.create(self, data_dict)



import collections
from functools import partial
from time import time


try:
    import cPickle as pickle
except ImportError:
    import pickle


def make_key(resource_name, resource_id=None):
    if not resource_id:
        return resource_name
    else:
        return '{0}:{1}'.format(resource_name, resource_id)


def update(d, u):
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def _items(mappingorseq):
    """Wrapper for efficient iteraiton over mappings represeted by dicts or
    sequeunces::

        >>> for k, v in _items([i, i*i] for i in xrange(5)):
        ...     assert k*k == v

        >>> for k, v in _items({i: i*i} for i in xrange(5)):
        ...     assert k*k == v
    """
    return mappingorseq.iteritems() if hasattr(mappingorseq, 'iteritems') \
        else mappingorseq


class BaseCache(object):
    """Base class for the cache system. All the cache systems will implemtent
    this API or a superset of it.

    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`set`.
    """

    def __init__(self, default_timeout=300):
        self.default_timeout = default_timeout

    def get(self, resource_name, resource_id=None):
        return None

    def set(self, resource_name, data_dict):
        pass

    def set_many(self, resource_name, sequence, timeout=None):
        for data_dict in sequence:
            self.set(resource_name, data_dict, timeout)

    def get_many(self, resource_name, ids):
        func = partial(self.get, resource_name)
        return map(func, ids)

    def delete(self, resource_name, resource_id=None):
        pass

    def clear(self):
        pass

    def count(self):
        raise NotImplementedError


class NullCache(BaseCache):
    """
    A cache that doesn't cache.
    """


class SimpleCache(BaseCache):
    """Simply memory cache for single process environments. This class exists
    mainly for a development server and is not 100% thread safe.

    :param threshold: the maximum number of items the cache stores before
                      it starts evicting keys.
    """
    def __init__(self, threshold=500, default_timeout=300):
        BaseCache.__init__(self, default_timeout)
        self._cache = {}
        self._threshold = threshold

    def _prune(self):
        if len(self._cache) > self._threshold:
            now = time()
            # Prune expired keys, or every few keys.
            for idx, (key, (expires, _)) in enumerate(self._cache.items()):
                if expires <= now or idx % 3 == 0:
                    self._cache.pop(key, None)

    def get(self, resource_name, resource_id=None):
        key = make_key(resource_name, resource_id)
        expires, value = self._cache.get(key, (0, None))
        if expires > time():
            return pickle.loads(value)

    def set(self, resource_name, data_dict, timeout=None):
        key = make_key(resource_name, data_dict.get('id'))
        if timeout is None:
            timeout = self.default_timeout
        self._prune()

        self._cache[key] = (time() + timeout, pickle.dumps(data_dict,
                            pickle.HIGHEST_PROTOCOL))

    def delete(self, resource_name, resource_id=None):
        key = make_key(resource_name, resource_id)
        return self._cache.pop(key, None)

    def clear(self):
        self._cache.clear()

    def count(self):
        return len(self._cache)


class RedisCache(BaseCache):
    """Uses the Redis key-value store as a cache backend.
    """
    def __init__(self, host='localhost', port=6379, password=None,
                 db=0, default_timeout=300, key_prefix=None):
        BaseCache.__init__(self, default_timeout)
        self.key_prefix = key_prefix or ''
        try:
            import redis
        except ImportError:
            raise RuntimeError('no redis module found')
        self._client = redis.Redis(host=host, port=port, password=password, db=db)

    def load_object(self, value):
        """The reversal of `dump_object`. This might be called with None.
        """
        if value is None:
            return None
        return pickle.loads(value)

    def dump_object(self, value):
        return pickle.dumps(value)

    def get(self, resource_name, resource_id=None):
        key = make_key(resource_name, resource_id)
        return self.load_object(self._client.get(self.key_prefix + key))

    def set(self, resource_name, data_dict, timeout=None):
        key = make_key(resource_name, data_dict.get('id', None))
        if timeout is None:
            timeout = self.default_timeout
        data = self.dump_object(data_dict)
        return self._client.setex(self.key_prefix + key, data, timeout)

    def delete(self, resource_name, resource_id=None):
        key = make_key(resource_name, resource_id)
        return self._client.delete(self.key_prefix + key)

    def clear(self):
        return self._client.flushall()

    def info(self):
        return self._client.info()


class RedisHashCache(RedisCache):
    """Use the Redis store as a cache backend, and do our best to store related
    items within hashes, making the best attempt to store items as native Redis
    objects, rather than pickling data. Efficiency!
    """
    def load_fields(self, data_dict):
        """
        Take a flattened dump of key-value pairs, and create a nested object
        from them.
        """
        data = {}

        def _update_data(k, v):
            d = {}
            keys = k.split('::', 1)
            nested_key, inner_key, = keys + (2 - len(keys)) * [None]
            if inner_key:
                d[nested_key] = _update_data(inner_key, v)
            else:
                d[nested_key] = v
            return d

        for k, v in data_dict.iteritems():
            if '::' in k:
                data = update(data, _update_data(k, v))
            else:
                data[k] = v
        if not data:
            return None
        return data

    def dump_fields(self, data_dict, prefix=''):
        """
        Convert any type of dictionary into a flat set of key-value pairs.
        """
        data = {}
        for k, v in data_dict.iteritems():
            if isinstance(v, dict):
                data.update(self.dump_fields(v, prefix=prefix + '%s::' % k))
            else:
                data[prefix + str(k)] = v
        return data

    def get(self, resource_name, resource_id=None):
        key = make_key(resource_name, resource_id)
        return self.load_fields(self._client.hgetall(self.key_prefix + key))

    def set(self, resource_name, data_dict, timeout=None):
        """
        In order to store things in a hash efficiently, we'll need to flatten
        data dictionaries down to simple key value pairs. The toughest part will
        be a nested dict, but a __ convention for a nesting should be fairly
        straightforward to implement.
        """
        key = make_key(resource_name, data_dict.get('id', None))
        if timeout is None:
            timeout = self.default_timeout

        data = self.dump_fields(data_dict)
        res = self._client.hmset(self.key_prefix + key, data)
        self._client.expire(self.key_prefix + key, timeout)
        return res



from __future__ import unicode_literals

from ..base import Resource


class TourCategory(Resource):

    _resource_name = 'tour_categories'

    _as_is_fields = ['id', 'href', 'name']
    _resource_fields = [('category_type', 'TourCategory')]



from mock import patch
import unittest

from gapipy.client import Client
from gapipy.query import Query
from gapipy.resources.base import Resource
from gapipy.utils import get_available_resource_classes


class ClientTestCase(unittest.TestCase):

    def setUp(self):
        self.gapi = Client()

    def test_resource_querysets_available(self):
        for resource in [r._resource_name for r in get_available_resource_classes()]:
            self.assertTrue(hasattr(self.gapi, resource))
            self.assertIsInstance(getattr(self.gapi, resource), Query)

    def test_query_interface(self):
        resource_name = get_available_resource_classes()[0]._resource_name
        self.assertTrue(isinstance(self.gapi.query(resource_name), Query))

    def test_unavailable_resource_query(self):
        with self.assertRaises(AttributeError):
            self.gapi.query("fake_resource")

    def test_build_interface(self):
        class MockResource(Resource):
            _as_is_fields = ['id', 'foo']
            _resource_name = 'foo'
        self.gapi.foo = Query(self.gapi, MockResource)

        resource = self.gapi.build('foo', {'id': 1, 'foo': 'bar'})

        self.assertEquals(resource.id, 1)
        self.assertEquals(resource.foo, 'bar')

    @patch('gapipy.request.APIRequestor._request')
    def test_create_interface(self, mock_request):
        class MockResource(Resource):
            _as_is_fields = ['id', 'foo']
            _resource_name = 'foo'
        self.gapi.foo = Query(self.gapi, MockResource)

        # On create, a representation of the resource is created.
        mock_request.return_value = {
            'id': 1,
            'foo': 'bar',
        }

        # Create allows arbitrary data to be sent, even if it's not part of the
        # final resource.
        resource = self.gapi.create('foo', {'id': 1, 'foo': 'bar', 'context': 'abc'})
        self.assertEquals(resource.id, 1)



from __future__ import unicode_literals

from ...models import PriceBand

from ..base import Product


class Activity(Product):

    _resource_name = 'activities'
    _is_listable = False

    # TODO: Move `activity_dossier` to `_resource_fields` once the
    # ActivityDossier resource is implemented.
    _as_is_fields = [
        'id', 'href', 'availability', 'name', 'description', 'product_line',
        'sku', 'type', 'sub_type', 'requirements', 'activity_dossier',
    ]
    _date_fields = ['start_date', 'finish_date']
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _model_collection_fields = [('price_bands', PriceBand)]



from __future__ import unicode_literals

from ..base import Resource
from .country import Country
from .feature import Feature
from .state import State
from .timezone import Timezone


class Place(Resource):

    _resource_name = 'places'

    _as_is_fields = [
        'id', 'href', 'name', 'ascii_name', 'population', 'elevation',
        'latitude', 'longitude', 'bounding_box', 'alternate_names',
        'admin_divisions',
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _resource_fields = [
        ('country', Country),
        ('state', State),
        ('feature', Feature),
        ('timezone', Timezone),
    ]

    def __init__(self, *args, **kwargs):
        super(Place, self).__init__(*args, **kwargs)
        self._set_admin_divisions()

    def _set_admin_divisions(self):
        """Transform the raw json list of `admin_divisions` into a list of the
        corresponding Place (stub) instances.
        """
        if 'admin_divisions' in self._raw_data:
            raw_admin_divisions = self._raw_data['admin_divisions']
            admin_divisions = [self.__class__(d, stub=True) for d in raw_admin_divisions]
            self.admin_divisions = admin_divisions



from .add_on import AddOn
from .price_band import PriceBand, SeasonalPriceBand
from .base import BaseModel


class Room(BaseModel):
    _as_is_fields = ['availability', 'code', 'name']

    @property
    def _model_collection_fields(self):
        return [('price_bands', PriceBand)]

    def __repr__(self):
        return '<{0} ({1})>'.format(self.__class__.__name__, self.name)


class AccommodationRoom(Room):
    model_collection_fields = [('price_bands', SeasonalPriceBand)]

    @property
    def _as_is_fields(self):
        return super(AccommodationRoom, self)._as_is_fields + [
            'room_class',
            'max_nights',
            'min_nights',
        ]

    def __init__(self, data):
        super(AccommodationRoom, self).__init__(data)

        # `class` is a reserved word in python
        setattr(self, 'room_class', data['class'])


class DepartureRoom(Room):
    @property
    def _as_is_fields(self):
        return super(DepartureRoom, self)._as_is_fields + [
            'flags',
        ]

    @property
    def _model_collection_fields(self):
        return super(DepartureRoom, self)._model_fields + [
            ('add_ons', AddOn),
        ]



from __future__ import unicode_literals

from ..base import Resource
from .place import Place


class Continent(Resource):

    _resource_name = 'continents'

    _as_is_fields = [
        'id', 'href', 'name',
    ]
    _date_time_fields_utc = ['date_created', 'date_last_modified']
    _resource_fields = [
        ('place', Place),
    ]



import datetime
from decimal import Decimal

from gapipy import client as client_module
from gapipy.query import Query
from gapipy.utils import (
    get_resource_class_from_class_name,
    get_resource_class_from_resource_name,
)

DATE_FORMAT = '%Y-%m-%d'
DATE_TIME_UTC_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_TIME_LOCAL_FORMAT = "%Y-%m-%dT%H:%M:%S"


class BaseModel(object):
    """
    A base for any Resource or model type. The difference being a Resource
    points to a specific API resource while a model type might have a reference
    to a resource but is mostly nested type that is complex and has attributes
    adjusted for it.
    """
    _as_is_fields = []
    _date_fields = []
    _date_time_fields_utc = []
    _date_time_fields_local = []
    _model_fields = []
    _model_collection_fields = []
    _price_fields = []
    _resource_fields = []
    _resource_collection_fields = []

    def __init__(self, data, client=None):
        self._client = client or client_module.current_client
        self._raw_data = data
        self._fill_fields(data)

    def _fill_fields(self, data):
        self._raw_data = data
        first = lambda l: [pair[0] for pair in l]

        # Initially we populate base fields, as model/resource fields may rely
        # on these to be present.
        remaining_data = {}
        for field, value in data.iteritems():
            if field in self._as_is_fields:
                self._set_as_is_field(field, value)
            elif field in self._date_fields:
                self._set_date_field(field, value)
            elif field in self._date_time_fields_local:
                self._set_date_time_field_local(field, value)
            elif field in self._date_time_fields_utc:
                self._set_date_time_field_utc(field, value)
            elif field in self._price_fields:
                self._set_price_field(field, value)
            else:
                remaining_data[field] = value

        # Populate resource/model fields.
        for field, value in remaining_data.iteritems():
            if field in first(self._model_fields):
                self._set_model_field(field, value)
            elif field in first(self._model_collection_fields):
                self._set_model_collection_field(field, value)
            elif field in first(self._resource_fields):
                self._set_resource_field(field, value)
            elif field in first(self._resource_collection_fields):
                self._set_resource_collection_field(field, value)

    def _set_as_is_field(self, field, value):
        setattr(self, field, value)

    def _set_date_field(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_FORMAT)
        setattr(self, field, value)

    def _set_date_time_field_local(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_TIME_LOCAL_FORMAT)
        setattr(self, field, value)

    def _set_date_time_field_utc(self, field, value):
        if value:
            value = datetime.datetime.strptime(value, DATE_TIME_UTC_FORMAT)
        setattr(self, field, value)

    def _set_price_field(self, field, value):
        if value:
            value = Decimal(value)
        setattr(self, field, value)

    def _model_cls(self, field):
        # Find what model class this field is in reference to by plucking it
        # from the set of all fields that allow this type of definition.
        fields = (self._model_fields
                  + self._model_collection_fields
                  + self._resource_fields
                  + self._resource_collection_fields)

        model_cls = [cls for f, cls in fields if f == field][0]

        # FIXME: This will not work for the model_*_fields.
        if isinstance(model_cls, basestring):
            model_cls = get_resource_class_from_class_name(model_cls)
        return model_cls

    def _set_model_field(self, field, value):
        setattr(self, field, self._model_cls(field)(value))

    def _set_model_collection_field(self, field, value):
        model_cls = self._model_cls(field)
        items = [model_cls(m) for m in value]
        setattr(self, field, items)

    def _set_resource_field(self, field, value):
        stub = None

        if isinstance(value, list):
            stub = []
            for v in value:
                stub.append(self._model_cls(field)(v, stub=True))
        elif value:
            stub = self._model_cls(field)(value, stub=True)
        setattr(self, field, stub)

    def _set_resource_collection_field(self, field, value):
        is_parent_resource = getattr(self, '_is_parent_resource', None)
        if is_parent_resource:
            parent = (self._resource_name, self.id)
        else:
            parent = None
        raw_data = value
        query = Query(self._client, self._model_cls(field), parent=parent, raw_data=raw_data)
        setattr(self, field, query)

    def _allowed_fields(self):
        first = lambda pair: pair[0]
        return (
            self._as_is_fields
            + self._date_fields
            + self._date_time_fields_utc
            + self._date_time_fields_local
            + map(first, self._model_fields)
            + map(first, self._model_collection_fields)
            + self._price_fields
            + map(first, self._resource_fields)
            + map(first, self._resource_collection_fields)
        )

    def _convert_from_resource_type(self, key, value):
        # Convert instance values into serializable objects.
        if isinstance(value, BaseModel):
            return value.to_dict()
        elif isinstance(value, Query):
            return value._to_dict()
        elif isinstance(value, Decimal):
            return str(value)
        elif isinstance(value, datetime.datetime):
            if key in self._date_fields:
                return datetime.datetime.strftime(value, DATE_FORMAT)
            elif key in self._date_time_fields_utc:
                return datetime.datetime.strftime(value, DATE_TIME_UTC_FORMAT)
            elif key in self._date_time_fields_local:
                return datetime.datetime.strftime(value, DATE_TIME_LOCAL_FORMAT)
        else:
            return value

    def to_dict(self):
        properties = {k: v for k, v in self.__dict__.iteritems() if k in self._allowed_fields()}
        data = {}
        for key, value in properties.iteritems():
            if isinstance(value, (list, tuple)):
                data[key] = [self._convert_from_resource_type(key, a) for a in value]
            else:
                data[key] = self._convert_from_resource_type(key, value)
        return data


class RelatedResourceMixin(object):
    _related_resource_lookup = None

    def __init__(self, data):
        super(RelatedResourceMixin, self).__init__(data)

        assert self._related_resource_lookup is not None

        resource_cls = get_resource_class_from_resource_name(
            getattr(self, self._related_resource_lookup))
        resource = resource_cls({'id': self.id, 'href': self.href}, stub=True)
        setattr(self, 'resource', resource)



"""These tests mostly ensure that any change in the shape of the data
returned by the API is detected.
"""

from unittest import TestCase

from nose.plugins.attrib import attr
from nose_parameterized import parameterized

from gapipy.client import Client
from gapipy.resources import (
    available_public_resources,
    Accommodation, Activity, Country, Departure, Language, Promotion,
    SingleSupplement, State, Tour, TourCategory, TourDossier, Transport,
    Feature, FeatureCategory, Timezone, Place, Airport, Continent,
)


@attr('integration')
class TourTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tour = Client().tours.get(21715)

    def setUp(self):
        self.tour = TourTestCase.tour

    def test_get_itinerary(self):
        itin = self.tour.get_brief_itinerary()
        self.assertIsInstance(itin, list)


@attr('integration')
class TourDossierTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dossier = Client().tour_dossiers.get(21715)

    def setUp(self):
        self.dossier = TourDossierTestCase.dossier

    def test_get_itinerary(self):
        itin = self.dossier.get_detailed_itinerary()
        self.assertIsInstance(itin, list)

    def test_get_image_url(self):
        url = self.dossier.get_map_url()
        self.assertIsInstance(url, basestring)

    def test_get_countries(self):
        countries = self.dossier.get_visited_countries()
        self.assertIsInstance(countries, list)

    def test_get_trip_detail(self):
        detail = self.dossier.get_trip_detail('Max Pax')
        self.assertIsInstance(detail, basestring)


@attr('integration')
class LiveAPITestCase(TestCase):

    resources = [
        (Accommodation, 54),
        (Activity, 4486),
        (Country, 'CA'),
        (Departure, 419504),
        (Language, 'aaa'),
        (Promotion, 18336),
        (SingleSupplement, 'T419504'),
        (State, 'CA-ON'),
        (Tour, 21715),
        (TourCategory, 15),
        (TourDossier, 21715),
        (Transport, 298),
        (Feature, 214),
        (FeatureCategory, 4),
        (Timezone, 4),
        (Place, 2039527),
        (Airport, 25553),
        (Continent, 'NA'),
    ]

    def test_all_public_resources_are_tested(self):
        """This a bit meta, this is a test to ensure that all available
        (public) resources are included in this integration test case.
        """
        tested_resources = [r.__name__ for (r, pk) in self.resources]
        self.assertEqual(sorted(available_public_resources), sorted(tested_resources))

    @parameterized.expand(resources)
    def test_fetch_resource(self, resource, resource_id):

        # Test if a resource can be parsed properly. This is useful to catch
        # changes in the API
        api = Client()
        query = getattr(api, resource._resource_name)
        instance = query.get(resource_id)

        # Test that we take into account all of the fields of the json data
        # returned by the API (this only checks for the top-level fields, not
        # those of nested objects)
        self.assertEquals(
            sorted(instance._allowed_fields()),
            sorted(instance._raw_data.keys())
        )
